@@ -105,7 +105,7 @@ __isl_null isl_ast_print_options *isl_ast_print_options_free(
 
 /* Set the print_user callback of "options" to "print_user".
  *
- * If this callback is set, then it used to print user nodes in the AST.
+ * If this callback is set, then it is used to print user nodes in the AST.
  * Otherwise, the expression associated to the user node is printed.
  */
 __isl_give isl_ast_print_options *isl_ast_print_options_set_print_user(
@@ -1018,6 +1018,11 @@ __isl_give isl_ast_node *isl_ast_node_copy(__isl_keep isl_ast_node *node)
 	return node;
 }
 
+/* Return a fresh copy of "node".
+ *
+ * In the case of a degenerate for node, take into account
+ * that "cond" and "inc" are NULL.
+ */
 __isl_give isl_ast_node *isl_ast_node_dup(__isl_keep isl_ast_node *node)
 {
 	isl_ast_node *dup;
@@ -1039,13 +1044,17 @@ __isl_give isl_ast_node *isl_ast_node_dup(__isl_keep isl_ast_node *node)
 			return isl_ast_node_free(dup);
 		break;
 	case isl_ast_node_for:
+		dup->u.f.degenerate = node->u.f.degenerate;
 		dup->u.f.iterator = isl_ast_expr_copy(node->u.f.iterator);
 		dup->u.f.init = isl_ast_expr_copy(node->u.f.init);
+		dup->u.f.body = isl_ast_node_copy(node->u.f.body);
+		if (!dup->u.f.iterator || !dup->u.f.init || !dup->u.f.body)
+			return isl_ast_node_free(dup);
+		if (node->u.f.degenerate)
+			break;
 		dup->u.f.cond = isl_ast_expr_copy(node->u.f.cond);
 		dup->u.f.inc = isl_ast_expr_copy(node->u.f.inc);
-		dup->u.f.body = isl_ast_node_copy(node->u.f.body);
-		if (!dup->u.f.iterator || !dup->u.f.init || !dup->u.f.cond ||
-		    !dup->u.f.inc || !dup->u.f.body)
+		if (!dup->u.f.cond || !dup->u.f.inc)
 			return isl_ast_node_free(dup);
 		break;
 	case isl_ast_node_block:
@@ -1068,6 +1077,12 @@ __isl_give isl_ast_node *isl_ast_node_dup(__isl_keep isl_ast_node *node)
 		break;
 	}
 
+	if (!node->annotation)
+		return dup;
+	dup->annotation = isl_id_copy(node->annotation);
+	if (!dup->annotation)
+		return isl_ast_node_free(dup);
+
 	return dup;
 }
 
@@ -1803,7 +1818,7 @@ __isl_give isl_printer *isl_ast_op_type_set_print_name(
  *
  * If there is a user-specified name in an isl_ast_expr_op_names note
  * associated to "p", then return that.
- * Otherwise, return the default name in op_str.
+ * Otherwise, return the default name in op_str_c.
  */
 static const char *get_op_str_c(__isl_keep isl_printer *p,
 	enum isl_ast_expr_op_type type)
@@ -2527,13 +2542,28 @@ error:
 }
 
 /* Print "node" to "p".
+ *
+ * "node" is assumed to be either the outermost node in an AST or
+ * a node that is known not to be a block.
+ * If "node" is a block (and is therefore outermost) and
+ * if the ast_print_outermost_block options is not set,
+ * then act as if the printing occurs inside a block, such
+ * that no "extra" block will get printed.
  */
 __isl_give isl_printer *isl_ast_node_print(__isl_keep isl_ast_node *node,
 	__isl_take isl_printer *p, __isl_take isl_ast_print_options *options)
 {
+	int in_block = 0;
+
 	if (!options || !node)
 		goto error;
-	p = print_ast_node_c(p, node, options, 0, 0);
+	if (node->type == isl_ast_node_block) {
+		isl_ctx *ctx;
+
+		ctx = isl_ast_node_get_ctx(node);
+		in_block = !isl_options_get_ast_print_outermost_block(ctx);
+	}
+	p = print_ast_node_c(p, node, options, in_block, 0);
 	isl_ast_print_options_free(options);
 	return p;
 error:
@@ -2588,10 +2618,10 @@ __isl_give isl_printer *isl_ast_node_list_print(
 	return p;
 }
 
-#define ISL_AST_MACRO_FLOORD	(1 << 0)
+#define ISL_AST_MACRO_FDIV_Q	(1 << 0)
 #define ISL_AST_MACRO_MIN	(1 << 1)
 #define ISL_AST_MACRO_MAX	(1 << 2)
-#define ISL_AST_MACRO_ALL	(ISL_AST_MACRO_FLOORD | \
+#define ISL_AST_MACRO_ALL	(ISL_AST_MACRO_FDIV_Q | \
 				 ISL_AST_MACRO_MIN | \
 				 ISL_AST_MACRO_MAX)
 
@@ -2613,7 +2643,7 @@ static int ast_expr_required_macros(__isl_keep isl_ast_expr *expr, int macros)
 	if (expr->u.op.op == isl_ast_expr_op_max)
 		macros |= ISL_AST_MACRO_MAX;
 	if (expr->u.op.op == isl_ast_expr_op_fdiv_q)
-		macros |= ISL_AST_MACRO_FLOORD;
+		macros |= ISL_AST_MACRO_FDIV_Q;
 
 	for (i = 0; i < expr->u.op.n_arg; ++i)
 		macros = ast_expr_required_macros(expr->u.op.args[i], macros);
@@ -2827,7 +2857,7 @@ static isl_stat foreach_ast_expr_op_type(int macros,
 		return isl_stat_error;
 	if (macros & ISL_AST_MACRO_MAX && fn(isl_ast_expr_op_max, user) < 0)
 		return isl_stat_error;
-	if (macros & ISL_AST_MACRO_FLOORD &&
+	if (macros & ISL_AST_MACRO_FDIV_Q &&
 	    fn(isl_ast_expr_op_fdiv_q, user) < 0)
 		return isl_stat_error;
 
