@@ -648,9 +648,9 @@ __isl_give isl_map *isl_ast_build_get_schedule_map(
 /* Return the position of the dimension in build->domain for which
  * an AST node is currently being generated.
  */
-int isl_ast_build_get_depth(__isl_keep isl_ast_build *build)
+isl_size isl_ast_build_get_depth(__isl_keep isl_ast_build *build)
 {
-	return build ? build->depth : -1;
+	return build ? build->depth : isl_size_error;
 }
 
 /* Prepare for generating code for the next level.
@@ -1373,14 +1373,14 @@ __isl_give isl_multi_aff *isl_ast_build_get_stride_expansion(
 {
 	isl_space *space;
 	isl_multi_aff *ma;
-	int pos;
+	isl_size pos;
 	isl_aff *aff, *offset;
 	isl_val *stride;
 
-	if (!build)
+	pos = isl_ast_build_get_depth(build);
+	if (pos < 0)
 		return NULL;
 
-	pos = isl_ast_build_get_depth(build);
 	space = isl_ast_build_get_space(build, 1);
 	space = isl_space_map_from_set(space);
 	ma = isl_multi_aff_identity(space);
@@ -1438,16 +1438,16 @@ __isl_give isl_ast_build *isl_ast_build_include_stride(
 __isl_give isl_ast_build *isl_ast_build_detect_strides(
 	__isl_take isl_ast_build *build, __isl_take isl_set *set)
 {
-	int pos;
+	isl_size pos;
 	isl_bool no_stride;
 	isl_val *stride;
 	isl_aff *offset;
 	isl_stride_info *si;
 
-	if (!build)
+	pos = isl_ast_build_get_depth(build);
+	if (pos < 0)
 		goto error;
 
-	pos = isl_ast_build_get_depth(build);
 	si = isl_set_get_stride_info(set, pos);
 	stride = isl_stride_info_get_stride(si);
 	offset = isl_stride_info_get_offset(si);
@@ -1467,44 +1467,27 @@ error:
 	return NULL;
 }
 
-struct isl_ast_build_involves_data {
-	int depth;
-	int involves;
-};
-
-/* Check if "map" involves the input dimension data->depth.
+/* Does "map" not involve the input dimension data->depth?
  */
-static isl_stat involves_depth(__isl_take isl_map *map, void *user)
+static isl_bool free_of_depth(__isl_keep isl_map *map, void *user)
 {
-	struct isl_ast_build_involves_data *data = user;
+	int *depth = user;
 
-	data->involves = isl_map_involves_dims(map, isl_dim_in, data->depth, 1);
-	isl_map_free(map);
-
-	if (data->involves < 0 || data->involves)
-		return isl_stat_error;
-	return isl_stat_ok;
+	return isl_bool_not(isl_map_involves_dims(map, isl_dim_in, *depth, 1));
 }
 
 /* Do any options depend on the value of the dimension at the current depth?
  */
 int isl_ast_build_options_involve_depth(__isl_keep isl_ast_build *build)
 {
-	struct isl_ast_build_involves_data data;
+	isl_bool free;
 
 	if (!build)
 		return -1;
 
-	data.depth = build->depth;
-	data.involves = 0;
-
-	if (isl_union_map_foreach_map(build->options,
-					&involves_depth, &data) < 0) {
-		if (data.involves < 0 || !data.involves)
-			return -1;
-	}
-
-	return data.involves;
+	free = isl_union_map_every_map(build->options, &free_of_depth,
+					&build->depth);
+	return isl_bool_not(free);
 }
 
 /* Construct the map
@@ -1966,7 +1949,7 @@ isl_bool isl_ast_build_has_stride(__isl_keep isl_ast_build *build, int pos)
  *
  *	f + s a
  *
- * with a an integer, return s through *stride.
+ * with a an integer, return s.
  */
 __isl_give isl_val *isl_ast_build_get_stride(__isl_keep isl_ast_build *build,
 	int pos)
