@@ -10,6 +10,7 @@
  * B.P. 105 - 78153 Le Chesnay, France
  */
 
+#include <isl/id.h>
 #include <isl/space.h>
 #include <isl/constraint.h>
 #include <isl/ilp.h>
@@ -98,7 +99,7 @@ static int is_non_neg_after_stealing(__isl_keep isl_aff *aff,
 	return non_neg;
 }
 
-/* Given the numerator "aff' of the argument of an integer division
+/* Given the numerator "aff" of the argument of an integer division
  * with denominator "d", steal part of the constant term of
  * the expression in which the integer division appears to make it
  * non-negative over data->build->domain.
@@ -114,7 +115,7 @@ static int is_non_neg_after_stealing(__isl_keep isl_aff *aff,
  * That is, compute the minimal value "m" of "aff" over
  * data->build->domain and take
  *
- *	s = ceil(m/d)
+ *	s = ceil(-m/d)
  *
  * such that
  *
@@ -187,14 +188,14 @@ static __isl_give isl_ast_expr *var_div(struct isl_ast_add_term_data *data,
 	isl_aff *aff;
 	isl_ast_expr *num, *den;
 	isl_val *d;
-	enum isl_ast_op_type type;
+	enum isl_ast_expr_op_type type;
 
 	aff = isl_local_space_get_div(ls, pos);
 	d = isl_aff_get_denominator_val(aff);
 	aff = isl_aff_scale_val(aff, isl_val_copy(d));
 	den = isl_ast_expr_from_val(isl_val_copy(d));
 
-	type = isl_ast_op_fdiv_q;
+	type = isl_ast_expr_op_fdiv_q;
 	if (isl_options_get_ast_build_prefer_pdiv(ctx)) {
 		int non_neg = isl_ast_build_aff_is_nonneg(data->build, aff);
 		if (non_neg >= 0 && !non_neg) {
@@ -216,7 +217,7 @@ static __isl_give isl_ast_expr *var_div(struct isl_ast_add_term_data *data,
 		if (non_neg < 0)
 			aff = isl_aff_free(aff);
 		else if (non_neg)
-			type = isl_ast_op_pdiv_q;
+			type = isl_ast_expr_op_pdiv_q;
 	}
 
 	isl_val_free(d);
@@ -296,11 +297,11 @@ error:
  * If expr2 is zero, we simply return expr1.
  * If expr1 is zero, we return
  *
- *	(isl_ast_op_minus, expr2)
+ *	(isl_ast_expr_op_minus, expr2)
  *
  * Otherwise, we return
  *
- *	(isl_ast_op_sub, expr1, expr2)
+ *	(isl_ast_expr_op_sub, expr1, expr2)
  */
 static __isl_give isl_ast_expr *ast_expr_sub(__isl_take isl_ast_expr *expr1,
 	__isl_take isl_ast_expr *expr2)
@@ -345,7 +346,7 @@ static __isl_give isl_ast_expr *isl_ast_expr_mod(__isl_keep isl_val *v,
 	expr = isl_ast_expr_from_aff(isl_aff_copy(aff), build);
 
 	c = isl_ast_expr_from_val(isl_val_copy(d));
-	expr = isl_ast_expr_alloc_binary(isl_ast_op_pdiv_r, expr, c);
+	expr = isl_ast_expr_alloc_binary(isl_ast_expr_op_pdiv_r, expr, c);
 
 	if (!isl_val_is_one(v)) {
 		c = isl_ast_expr_from_val(isl_val_copy(v));
@@ -360,11 +361,11 @@ static __isl_give isl_ast_expr *isl_ast_expr_mod(__isl_keep isl_val *v,
  * If v is 1, we simply return expr.
  * If v is -1, we return
  *
- *	(isl_ast_op_minus, expr)
+ *	(isl_ast_expr_op_minus, expr)
  *
  * Otherwise, we return
  *
- *	(isl_ast_op_mul, expr(v), expr)
+ *	(isl_ast_expr_op_mul, expr(v), expr)
  */
 static __isl_give isl_ast_expr *scale(__isl_take isl_ast_expr *expr,
 	__isl_take isl_val *v)
@@ -403,17 +404,17 @@ error:
  * multiplied by the absolute value of "*v".
  * If "*v" is negative, we create
  *
- *	(isl_ast_op_sub, expr, e)
+ *	(isl_ast_expr_op_sub, expr, e)
  *
  * except when expr is trivially zero, in which case we create
  *
- *	(isl_ast_op_minus, e)
+ *	(isl_ast_expr_op_minus, e)
  *
  * instead.
  *
  * If "*v" is positive, we simply create
  *
- *	(isl_ast_op_add, expr, e)
+ *	(isl_ast_expr_op_add, expr, e)
  *
  */
 static __isl_give isl_ast_expr *isl_ast_expr_add_term(
@@ -509,11 +510,41 @@ struct isl_extract_mod_data {
 	int sign;
 };
 
+/* Does
+ *
+ *	arg mod data->d
+ *
+ * represent (a special case of) a test for some linear expression
+ * being even?
+ *
+ * In particular, is it of the form
+ *
+ *	(lin - 1) mod 2
+ *
+ * ?
+ */
+static isl_bool is_even_test(struct isl_extract_mod_data *data,
+	__isl_keep isl_aff *arg)
+{
+	isl_bool res;
+	isl_val *cst;
+
+	res = isl_val_eq_si(data->d, 2);
+	if (res < 0 || !res)
+		return res;
+
+	cst = isl_aff_get_constant_val(arg);
+	res = isl_val_eq_si(cst, -1);
+	isl_val_free(cst);
+
+	return res;
+}
+
 /* Given that data->v * div_i in data->aff is equal to
  *
  *	f * (term - (arg mod d))
  *
- * with data->d * f = data->v, add
+ * with data->d * f = data->v and "arg" non-negative on data->build, add
  *
  *	f * term
  *
@@ -522,13 +553,38 @@ struct isl_extract_mod_data {
  *	abs(f) * (arg mod d)
  *
  * to data->neg or data->pos depending on the sign of -f.
+ *
+ * In the special case that "arg mod d" is of the form "(lin - 1) mod 2",
+ * with "lin" some linear expression, first replace
+ *
+ *	f * (term - ((lin - 1) mod 2))
+ *
+ * by
+ *
+ *	-f * (1 - term - (lin mod 2))
+ *
+ * These two are equal because
+ *
+ *	((lin - 1) mod 2) + (lin mod 2) = 1
+ *
+ * Also, if "lin - 1" is non-negative, then "lin" is non-negative too.
  */
 static int extract_term_and_mod(struct isl_extract_mod_data *data,
 	__isl_take isl_aff *term, __isl_take isl_aff *arg)
 {
+	isl_bool even;
 	isl_ast_expr *expr;
 	int s;
 
+	even = is_even_test(data, arg);
+	if (even < 0) {
+		arg = isl_aff_free(arg);
+	} else if (even) {
+		term = oppose_div_arg(term, isl_val_copy(data->d));
+		data->v = isl_val_neg(data->v);
+		arg = isl_aff_set_constant_si(arg, 0);
+	}
+
 	data->v = isl_val_div(data->v, isl_val_copy(data->d));
 	s = isl_val_sgn(data->v);
 	data->v = isl_val_abs(data->v);
@@ -678,11 +734,13 @@ static isl_stat check_parallel_or_opposite(__isl_take isl_constraint *c,
 	enum isl_dim_type c_type[2] = { isl_dim_param, isl_dim_set };
 	enum isl_dim_type a_type[2] = { isl_dim_param, isl_dim_in };
 	int i, t;
-	int n[2];
+	isl_size n[2];
 	int parallel = 1, opposite = 1;
 
 	for (t = 0; t < 2; ++t) {
 		n[t] = isl_constraint_dim(c, c_type[t]);
+		if (n[t] < 0)
+			return isl_stat_error;
 		for (i = 0; i < n[t]; ++i) {
 			int a, b;
 
@@ -796,7 +854,7 @@ static isl_stat check_parallel_or_opposite(__isl_take isl_constraint *c,
  * multiple of d to make it positive.
  *
  *
- * Note that the above is a only a very simple heuristic for finding an
+ * Note that the above is only a very simple heuristic for finding an
  * appropriate expression.  We could try a bit harder by also considering
  * sums of constraints that involve disjoint sets of variables or
  * we could consider arbitrary linear combinations of constraints,
@@ -823,12 +881,15 @@ static int try_extract_mod(struct isl_extract_mod_data *data)
 {
 	isl_basic_set *hull;
 	isl_val *v1, *v2;
-	int r, n;
+	isl_stat r;
+	isl_size n;
 
 	if (!data->build)
 		goto error;
 
 	n = isl_aff_dim(data->div, isl_dim_div);
+	if (n < 0)
+		goto error;
 
 	if (isl_aff_involves_dims(data->div, isl_dim_div, 0, n))
 		return extract_nonneg_mod(data);
@@ -900,7 +961,7 @@ error:
  *
  * Note that in order to represent "a mod m" as
  *
- *	(isl_ast_op_pdiv_r, a, m)
+ *	(isl_ast_expr_op_pdiv_r, a, m)
  *
  * we need to make sure that a is non-negative.
  * If not, we check if "-a + m - 1" is non-negative.
@@ -953,7 +1014,7 @@ static __isl_give isl_aff *extract_modulos(__isl_take isl_aff *aff,
 {
 	struct isl_extract_mod_data data = { build, aff, *pos, *neg };
 	isl_ctx *ctx;
-	int n;
+	isl_size n;
 
 	if (!aff)
 		return NULL;
@@ -963,6 +1024,8 @@ static __isl_give isl_aff *extract_modulos(__isl_take isl_aff *aff,
 		return aff;
 
 	n = isl_aff_dim(data.aff, isl_dim_div);
+	if (n < 0)
+		return isl_aff_free(aff);
 	for (data.i = 0; data.i < n; ++data.i) {
 		data.v = isl_aff_get_coefficient_val(data.aff,
 							isl_dim_div, data.i);
@@ -999,7 +1062,8 @@ static __isl_give isl_aff *extract_modulos(__isl_take isl_aff *aff,
 static __isl_give isl_aff *extract_rational(__isl_take isl_aff *aff,
 	__isl_keep isl_ast_expr **expr, __isl_keep isl_ast_build *build)
 {
-	int i, j, n;
+	int i, j;
+	isl_size n;
 	isl_aff *rat = NULL;
 	isl_local_space *ls = NULL;
 	isl_ast_expr *rat_expr;
@@ -1024,6 +1088,8 @@ static __isl_give isl_aff *extract_rational(__isl_take isl_aff *aff,
 
 	for (i = 0; i < 3; ++i) {
 		n = isl_aff_dim(aff, t[i]);
+		if (n < 0)
+			goto error;
 		for (j = 0; j < n; ++j) {
 			isl_aff *rat_j;
 
@@ -1069,7 +1135,7 @@ error:
 	return NULL;
 }
 
-/* Construct an isl_ast_expr that evaluates the affine expression "aff",
+/* Construct an isl_ast_expr that evaluates the affine expression "aff".
  * The result is simplified in terms of build->domain.
  *
  * We first extract hidden modulo computations from the affine expression
@@ -1081,7 +1147,7 @@ __isl_give isl_ast_expr *isl_ast_expr_from_aff(__isl_take isl_aff *aff,
 	__isl_keep isl_ast_build *build)
 {
 	int i, j;
-	int n;
+	isl_size n;
 	isl_val *v;
 	isl_ctx *ctx = isl_aff_get_ctx(aff);
 	isl_ast_expr *expr, *expr_neg;
@@ -1107,6 +1173,8 @@ __isl_give isl_ast_expr *isl_ast_expr_from_aff(__isl_take isl_aff *aff,
 	data.cst = isl_aff_get_constant_val(aff);
 	for (i = 0; i < 3; ++i) {
 		n = isl_aff_dim(aff, t[i]);
+		if (n < 0)
+			expr = isl_ast_expr_free(expr);
 		for (j = 0; j < n; ++j) {
 			v = isl_aff_get_coefficient_val(aff, t[i], j);
 			if (!v)
@@ -1143,7 +1211,9 @@ static __isl_give isl_ast_expr *add_signed_terms(__isl_take isl_ast_expr *expr,
 	ls = isl_aff_get_domain_local_space(aff);
 
 	for (i = 0; i < 3; ++i) {
-		int n = isl_aff_dim(aff, t[i]);
+		isl_size n = isl_aff_dim(aff, t[i]);
+		if (n < 0)
+			expr = isl_ast_expr_free(expr);
 		for (j = 0; j < n; ++j) {
 			v = isl_aff_get_coefficient_val(aff, t[i], j);
 			if (sign * isl_val_sgn(v) <= 0) {
@@ -1234,24 +1304,26 @@ static int is_stride_constraint(__isl_keep isl_aff *aff, int pos)
 
 /* Are all coefficients of "aff" (zero or) negative?
  */
-static int all_negative_coefficients(__isl_keep isl_aff *aff)
+static isl_bool all_negative_coefficients(__isl_keep isl_aff *aff)
 {
-	int i, n;
-
-	if (!aff)
-		return 0;
+	int i;
+	isl_size n;
 
 	n = isl_aff_dim(aff, isl_dim_param);
+	if (n < 0)
+		return isl_bool_error;
 	for (i = 0; i < n; ++i)
 		if (isl_aff_coefficient_sgn(aff, isl_dim_param, i) > 0)
-			return 0;
+			return isl_bool_false;
 
 	n = isl_aff_dim(aff, isl_dim_in);
+	if (n < 0)
+		return isl_bool_error;
 	for (i = 0; i < n; ++i)
 		if (isl_aff_coefficient_sgn(aff, isl_dim_in, i) > 0)
-			return 0;
+			return isl_bool_false;
 
-	return 1;
+	return isl_bool_true;
 }
 
 /* Give an equality of the form
@@ -1265,17 +1337,20 @@ static int all_negative_coefficients(__isl_keep isl_aff *aff)
  * with the integer division "pos" equal to floor(e/d),
  * construct the AST expression
  *
- *	(isl_ast_op_eq, (isl_ast_op_zdiv_r, expr(e), expr(d)), expr(0))
+ *	(isl_ast_expr_op_eq,
+ *		(isl_ast_expr_op_zdiv_r, expr(e), expr(d)), expr(0))
  *
  * If e only has negative coefficients, then construct
  *
- *	(isl_ast_op_eq, (isl_ast_op_zdiv_r, expr(-e), expr(d)), expr(0))
+ *	(isl_ast_expr_op_eq,
+ *		(isl_ast_expr_op_zdiv_r, expr(-e), expr(d)), expr(0))
  *
  * instead.
  */
 static __isl_give isl_ast_expr *extract_stride_constraint(
 	__isl_take isl_aff *aff, int pos, __isl_keep isl_ast_build *build)
 {
+	isl_bool all_neg;
 	isl_ctx *ctx;
 	isl_val *c;
 	isl_ast_expr *expr, *cst;
@@ -1288,20 +1363,23 @@ static __isl_give isl_ast_expr *extract_stride_constraint(
 	c = isl_aff_get_coefficient_val(aff, isl_dim_div, pos);
 	aff = isl_aff_set_coefficient_si(aff, isl_dim_div, pos, 0);
 
-	if (all_negative_coefficients(aff))
+	all_neg = all_negative_coefficients(aff);
+	if (all_neg < 0)
+		aff = isl_aff_free(aff);
+	else if (all_neg)
 		aff = isl_aff_neg(aff);
 
 	cst = isl_ast_expr_from_val(isl_val_abs(c));
 	expr = isl_ast_expr_from_aff(aff, build);
 
-	expr = isl_ast_expr_alloc_binary(isl_ast_op_zdiv_r, expr, cst);
+	expr = isl_ast_expr_alloc_binary(isl_ast_expr_op_zdiv_r, expr, cst);
 	cst = isl_ast_expr_alloc_int_si(ctx, 0);
-	expr = isl_ast_expr_alloc_binary(isl_ast_op_eq, expr, cst);
+	expr = isl_ast_expr_alloc_binary(isl_ast_expr_op_eq, expr, cst);
 
 	return expr;
 }
 
-/* Construct an isl_ast_expr that evaluates the condition "constraint",
+/* Construct an isl_ast_expr that evaluates the condition "constraint".
  * The result is simplified in terms of build->domain.
  *
  * We first check if the constraint is an equality of the form
@@ -1314,7 +1392,8 @@ static __isl_give isl_ast_expr *extract_stride_constraint(
  *
  * If so, we convert it to
  *
- *	(isl_ast_op_eq, (isl_ast_op_zdiv_r, expr(e), expr(d)), expr(0))
+ *	(isl_ast_expr_op_eq,
+ *		(isl_ast_expr_op_zdiv_r, expr(e), expr(d)), expr(0))
  *
  * Otherwise, let the constraint by either "a >= 0" or "a == 0".
  * We first extract hidden modulo computations from "a"
@@ -1323,31 +1402,32 @@ static __isl_give isl_ast_expr *extract_stride_constraint(
  *
  * The result is then of the form
  *
- *	(isl_ast_op_ge, expr(pos), expr(-neg)))
+ *	(isl_ast_expr_op_ge, expr(pos), expr(-neg)))
  *
  * or
  *
- *	(isl_ast_op_eq, expr(pos), expr(-neg)))
+ *	(isl_ast_expr_op_eq, expr(pos), expr(-neg)))
  *
  * However, if the first expression is an integer constant (and the second
  * is not), then we swap the two expressions.  This ensures that we construct,
  * e.g., "i <= 5" rather than "5 >= i".
  *
- * Furthermore, is there are no terms with positive coefficients (or no terms
+ * Furthermore, if there are no terms with positive coefficients (or no terms
  * with negative coefficients), then the constant term is added to "pos"
  * (or "neg"), ignoring the sign of the constant term.
  */
 static __isl_give isl_ast_expr *isl_ast_expr_from_constraint(
 	__isl_take isl_constraint *constraint, __isl_keep isl_ast_build *build)
 {
-	int i, n;
+	int i;
+	isl_size n;
 	isl_ctx *ctx;
 	isl_ast_expr *expr_pos;
 	isl_ast_expr *expr_neg;
 	isl_ast_expr *expr;
 	isl_aff *aff;
 	int eq;
-	enum isl_ast_op_type type;
+	enum isl_ast_expr_op_type type;
 	struct isl_ast_add_term_data data;
 
 	if (!constraint)
@@ -1358,6 +1438,8 @@ static __isl_give isl_ast_expr *isl_ast_expr_from_constraint(
 	isl_constraint_free(constraint);
 
 	n = isl_aff_dim(aff, isl_dim_div);
+	if (n < 0)
+		aff = isl_aff_free(aff);
 	if (eq && n > 0)
 		for (i = 0; i < n; ++i) {
 			int is_stride;
@@ -1390,10 +1472,10 @@ static __isl_give isl_ast_expr *isl_ast_expr_from_constraint(
 
 	if (isl_ast_expr_get_type(expr_pos) == isl_ast_expr_int &&
 	    isl_ast_expr_get_type(expr_neg) != isl_ast_expr_int) {
-		type = eq ? isl_ast_op_eq : isl_ast_op_le;
+		type = eq ? isl_ast_expr_op_eq : isl_ast_expr_op_le;
 		expr = isl_ast_expr_alloc_binary(type, expr_neg, expr_pos);
 	} else {
-		type = eq ? isl_ast_op_eq : isl_ast_op_ge;
+		type = eq ? isl_ast_expr_op_eq : isl_ast_expr_op_ge;
 		expr = isl_ast_expr_alloc_binary(type, expr_pos, expr_neg);
 	}
 
@@ -1423,7 +1505,7 @@ static int cmp_constraint(__isl_keep isl_constraint *a,
 /* Construct an isl_ast_expr that evaluates the conditions defining "bset".
  * The result is simplified in terms of build->domain.
  *
- * If "bset" is not bounded by any constraint, then we contruct
+ * If "bset" is not bounded by any constraint, then we construct
  * the expression "1", i.e., "true".
  *
  * Otherwise, we sort the constraints, putting constraints that involve
@@ -1443,7 +1525,8 @@ static int cmp_constraint(__isl_keep isl_constraint *a,
 __isl_give isl_ast_expr *isl_ast_build_expr_from_basic_set(
 	 __isl_keep isl_ast_build *build, __isl_take isl_basic_set *bset)
 {
-	int i, n;
+	int i;
+	isl_size n;
 	isl_constraint *c;
 	isl_constraint_list *list;
 	isl_ast_expr *res;
@@ -1452,9 +1535,9 @@ __isl_give isl_ast_expr *isl_ast_build_expr_from_basic_set(
 	list = isl_basic_set_get_constraint_list(bset);
 	isl_basic_set_free(bset);
 	list = isl_constraint_list_sort(list, &cmp_constraint, NULL);
-	if (!list)
-		return NULL;
 	n = isl_constraint_list_n_constraint(list);
+	if (n < 0)
+		build = NULL;
 	if (n == 0) {
 		isl_ctx *ctx = isl_constraint_list_get_ctx(list);
 		isl_constraint_list_free(list);
@@ -1509,7 +1592,8 @@ __isl_give isl_ast_expr *isl_ast_build_expr_from_basic_set(
 __isl_give isl_ast_expr *isl_ast_build_expr_from_set_internal(
 	__isl_keep isl_ast_build *build, __isl_take isl_set *set)
 {
-	int i, n;
+	int i;
+	isl_size n;
 	isl_basic_set *bset;
 	isl_basic_set_list *list;
 	isl_set *domain;
@@ -1518,9 +1602,9 @@ __isl_give isl_ast_expr *isl_ast_build_expr_from_set_internal(
 	list = isl_set_get_basic_set_list(set);
 	isl_set_free(set);
 
-	if (!list)
-		return NULL;
 	n = isl_basic_set_list_n_basic_set(list);
+	if (n < 0)
+		build = NULL;
 	if (n == 0) {
 		isl_ctx *ctx = isl_ast_build_get_ctx(build);
 		isl_basic_set_list_free(list);
@@ -1569,7 +1653,12 @@ __isl_give isl_ast_expr *isl_ast_build_expr_from_set_internal(
 __isl_give isl_ast_expr *isl_ast_build_expr_from_set(
 	__isl_keep isl_ast_build *build, __isl_take isl_set *set)
 {
-	if (isl_ast_build_need_schedule_map(build)) {
+	isl_bool needs_map;
+
+	needs_map = isl_ast_build_need_schedule_map(build);
+	if (needs_map < 0) {
+		set = isl_set_free(set);
+	} else if (needs_map) {
 		isl_multi_aff *ma;
 		ma = isl_ast_build_get_schedule_map_multi_aff(build);
 		set = isl_set_preimage_multi_aff(set, ma);
@@ -1645,11 +1734,13 @@ struct isl_from_pw_aff_data {
 static isl_stat isl_from_pw_aff_data_init(struct isl_from_pw_aff_data *data,
 	__isl_keep isl_ast_build *build, __isl_keep isl_pw_aff *pa)
 {
-	int n;
+	isl_size n;
 	isl_ctx *ctx;
 
 	ctx = isl_pw_aff_get_ctx(pa);
 	n = isl_pw_aff_n_piece(pa);
+	if (n < 0)
+		return isl_stat_error;
 	if (n == 0)
 		isl_die(ctx, isl_error_invalid,
 			"cannot handle void expression", return isl_stat_error);
@@ -1770,10 +1861,11 @@ static __isl_give isl_ast_expr *ast_expr_from_aff_list(
 	__isl_take isl_aff_list *list, enum isl_from_pw_aff_state state,
 	__isl_keep isl_ast_build *build)
 {
-	int i, n;
+	int i;
+	isl_size n;
 	isl_aff *aff;
-	isl_ast_expr *expr;
-	enum isl_ast_op_type op_type;
+	isl_ast_expr *expr = NULL;
+	enum isl_ast_expr_op_type op_type;
 
 	if (state == isl_state_single) {
 		aff = isl_aff_list_get_aff(list, 0);
@@ -1781,7 +1873,10 @@ static __isl_give isl_ast_expr *ast_expr_from_aff_list(
 		return isl_ast_expr_from_aff(aff, build);
 	}
 	n = isl_aff_list_n_aff(list);
-	op_type = state == isl_state_min ? isl_ast_op_min : isl_ast_op_max;
+	if (n < 0)
+		goto error;
+	op_type = state == isl_state_min ? isl_ast_expr_op_min
+					 : isl_ast_expr_op_max;
 	expr = isl_ast_expr_alloc_op(isl_ast_build_get_ctx(build), op_type, n);
 	if (!expr)
 		goto error;
@@ -1828,7 +1923,7 @@ static isl_ast_expr **add_intermediate_piece(struct isl_from_pw_aff_data *data,
 	set = data->p[pos].set;
 	data->p[pos].set = NULL;
 	ctx = isl_ast_build_get_ctx(data->build);
-	ternary = isl_ast_expr_alloc_op(ctx, isl_ast_op_select, 3);
+	ternary = isl_ast_expr_alloc_op(ctx, isl_ast_expr_op_select, 3);
 	gist = isl_set_gist(isl_set_copy(set), isl_set_copy(data->dom));
 	arg = isl_ast_build_expr_from_set_internal(data->build, gist);
 	ternary = isl_ast_expr_set_op_arg(ternary, 0, arg);
@@ -1891,7 +1986,7 @@ static int sort_pieces_cmp(const void *p1, const void *p2, void *arg)
 {
 	const struct isl_from_pw_aff_piece *piece1 = p1;
 	const struct isl_from_pw_aff_piece *piece2 = p2;
-	int n1, n2;
+	isl_size n1, n2;
 
 	n1 = isl_set_n_basic_set(piece1->set);
 	n2 = isl_set_n_basic_set(piece2->set);
@@ -1982,10 +2077,14 @@ static isl_bool aff_is_rational(__isl_keep isl_aff *aff)
  */
 static isl_bool is_single_rational_aff(__isl_keep isl_aff_list *list)
 {
+	isl_size n;
 	isl_bool rational;
 	isl_aff *aff;
 
-	if (isl_aff_list_n_aff(list) != 1)
+	n = isl_aff_list_n_aff(list);
+	if (n < 0)
+		return isl_bool_error;
+	if (n != 1)
 		return isl_bool_false;
 	aff = isl_aff_list_get_aff(list, 0);
 	rational = aff_is_rational(aff);
@@ -2018,7 +2117,8 @@ static isl_bool extends(struct isl_from_pw_aff_data *data,
 	__isl_give isl_basic_set *(*test)(__isl_take isl_aff *aff1,
 		__isl_take isl_aff *aff2))
 {
-	int i, n;
+	int i;
+	isl_size n;
 	isl_bool is_rational;
 	isl_ctx *ctx;
 	isl_set *dom;
@@ -2033,10 +2133,13 @@ static isl_bool extends(struct isl_from_pw_aff_data *data,
 	if (!isl_options_get_ast_build_detect_min_max(ctx))
 		return isl_bool_false;
 
+	n = isl_set_list_n_set(data->p[data->n].set_list);
+	if (n < 0)
+		return isl_bool_error;
+
 	dom = isl_ast_build_get_domain(data->build);
 	set = isl_set_intersect(dom, isl_set_copy(set));
 
-	n = isl_set_list_n_set(data->p[data->n].set_list);
 	for (i = 0; i < n ; ++i) {
 		isl_aff *aff_i;
 		isl_set *valid;
@@ -2205,8 +2308,12 @@ __isl_give isl_ast_expr *isl_ast_build_expr_from_pw_aff(
 	__isl_keep isl_ast_build *build, __isl_take isl_pw_aff *pa)
 {
 	isl_ast_expr *expr;
+	isl_bool needs_map;
 
-	if (isl_ast_build_need_schedule_map(build)) {
+	needs_map = isl_ast_build_need_schedule_map(build);
+	if (needs_map < 0) {
+		pa = isl_pw_aff_free(pa);
+	} else if (needs_map) {
 		isl_multi_aff *ma;
 		ma = isl_ast_build_get_schedule_map_multi_aff(build);
 		pa = isl_pw_aff_pullback_multi_aff(pa, ma);
@@ -2223,9 +2330,12 @@ __isl_give isl_ast_expr *isl_ast_build_expr_from_pw_aff(
 static __isl_give isl_multi_pw_aff *set_iterator_names(
 	__isl_keep isl_ast_build *build, __isl_take isl_multi_pw_aff *mpa)
 {
-	int i, n;
+	int i;
+	isl_size n;
 
 	n = isl_multi_pw_aff_dim(mpa, isl_dim_in);
+	if (n < 0)
+		return isl_multi_pw_aff_free(mpa);
 	for (i = 0; i < n; ++i) {
 		isl_id *id;
 
@@ -2242,17 +2352,18 @@ static __isl_give isl_multi_pw_aff *set_iterator_names(
  * with arguments/indices specified by "mpa".
  */
 static __isl_give isl_ast_expr *isl_ast_build_with_arguments(
-	__isl_keep isl_ast_build *build, enum isl_ast_op_type type,
+	__isl_keep isl_ast_build *build, enum isl_ast_expr_op_type type,
 	__isl_take isl_ast_expr *arg0, __isl_take isl_multi_pw_aff *mpa)
 {
-	int i, n;
+	int i;
+	isl_size n;
 	isl_ctx *ctx;
 	isl_ast_expr *expr;
 
 	ctx = isl_ast_build_get_ctx(build);
 
 	n = isl_multi_pw_aff_dim(mpa, isl_dim_out);
-	expr = isl_ast_expr_alloc_op(ctx, type, 1 + n);
+	expr = n >= 0 ? isl_ast_expr_alloc_op(ctx, type, 1 + n) : NULL;
 	expr = isl_ast_expr_set_op_arg(expr, 0, arg0);
 	for (i = 0; i < n; ++i) {
 		isl_pw_aff *pa;
@@ -2268,7 +2379,7 @@ static __isl_give isl_ast_expr *isl_ast_build_with_arguments(
 }
 
 static __isl_give isl_ast_expr *isl_ast_build_from_multi_pw_aff_internal(
-	__isl_keep isl_ast_build *build, enum isl_ast_op_type type,
+	__isl_keep isl_ast_build *build, enum isl_ast_expr_op_type type,
 	__isl_take isl_multi_pw_aff *mpa);
 
 /* Construct an isl_ast_expr that accesses the member specified by "mpa".
@@ -2285,7 +2396,7 @@ static __isl_give isl_ast_expr *isl_ast_build_from_multi_pw_aff_member(
 	isl_id *id;
 	isl_multi_pw_aff *domain;
 	isl_ast_expr *domain_expr, *expr;
-	enum isl_ast_op_type type = isl_ast_op_access;
+	enum isl_ast_expr_op_type type = isl_ast_expr_op_access;
 
 	domain = isl_multi_pw_aff_copy(mpa);
 	domain = isl_multi_pw_aff_range_factor_domain(domain);
@@ -2297,7 +2408,8 @@ static __isl_give isl_ast_expr *isl_ast_build_from_multi_pw_aff_member(
 			"missing field name", goto error);
 	id = isl_multi_pw_aff_get_tuple_id(mpa, isl_dim_out);
 	expr = isl_ast_expr_from_id(id);
-	expr = isl_ast_expr_alloc_binary(isl_ast_op_member, domain_expr, expr);
+	expr = isl_ast_expr_alloc_binary(isl_ast_expr_op_member,
+					domain_expr, expr);
 	return isl_ast_build_with_arguments(build, type, expr, mpa);
 error:
 	isl_multi_pw_aff_free(mpa);
@@ -2315,7 +2427,7 @@ error:
  * The domain of "mpa" is assumed to live in the internal schedule domain.
  */
 static __isl_give isl_ast_expr *isl_ast_build_from_multi_pw_aff_internal(
-	__isl_keep isl_ast_build *build, enum isl_ast_op_type type,
+	__isl_keep isl_ast_build *build, enum isl_ast_expr_op_type type,
 	__isl_take isl_multi_pw_aff *mpa)
 {
 	isl_ctx *ctx;
@@ -2325,7 +2437,7 @@ static __isl_give isl_ast_expr *isl_ast_build_from_multi_pw_aff_internal(
 	if (!mpa)
 		goto error;
 
-	if (type == isl_ast_op_access &&
+	if (type == isl_ast_expr_op_access &&
 	    isl_multi_pw_aff_range_is_wrapping(mpa))
 		return isl_ast_build_from_multi_pw_aff_member(build, mpa);
 
@@ -2355,7 +2467,7 @@ error:
  * The domain of "pma" is assumed to live in the internal schedule domain.
  */
 static __isl_give isl_ast_expr *isl_ast_build_from_pw_multi_aff_internal(
-	__isl_keep isl_ast_build *build, enum isl_ast_op_type type,
+	__isl_keep isl_ast_build *build, enum isl_ast_expr_op_type type,
 	__isl_take isl_pw_multi_aff *pma)
 {
 	isl_multi_pw_aff *mpa;
@@ -2372,10 +2484,11 @@ static __isl_give isl_ast_expr *isl_ast_build_from_pw_multi_aff_internal(
  * The domain of "mpa" is assumed to live in the external schedule domain.
  */
 static __isl_give isl_ast_expr *isl_ast_build_from_multi_pw_aff(
-	__isl_keep isl_ast_build *build, enum isl_ast_op_type type,
+	__isl_keep isl_ast_build *build, enum isl_ast_expr_op_type type,
 	__isl_take isl_multi_pw_aff *mpa)
 {
-	int is_domain;
+	isl_bool is_domain;
+	isl_bool needs_map;
 	isl_ast_expr *expr;
 	isl_space *space_build, *space_mpa;
 
@@ -2391,7 +2504,10 @@ static __isl_give isl_ast_expr *isl_ast_build_from_multi_pw_aff(
 		isl_die(isl_ast_build_get_ctx(build), isl_error_invalid,
 			"spaces don't match", goto error);
 
-	if (isl_ast_build_need_schedule_map(build)) {
+	needs_map = isl_ast_build_need_schedule_map(build);
+	if (needs_map < 0)
+		goto error;
+	if (needs_map) {
 		isl_multi_aff *ma;
 		ma = isl_ast_build_get_schedule_map_multi_aff(build);
 		mpa = isl_multi_pw_aff_pullback_multi_aff(mpa, ma);
@@ -2413,7 +2529,8 @@ error:
 __isl_give isl_ast_expr *isl_ast_build_call_from_multi_pw_aff(
 	__isl_keep isl_ast_build *build, __isl_take isl_multi_pw_aff *mpa)
 {
-	return isl_ast_build_from_multi_pw_aff(build, isl_ast_op_call, mpa);
+	return isl_ast_build_from_multi_pw_aff(build,
+						isl_ast_expr_op_call, mpa);
 }
 
 /* Construct an isl_ast_expr that accesses the array element specified by "mpa".
@@ -2425,7 +2542,8 @@ __isl_give isl_ast_expr *isl_ast_build_call_from_multi_pw_aff(
 __isl_give isl_ast_expr *isl_ast_build_access_from_multi_pw_aff(
 	__isl_keep isl_ast_build *build, __isl_take isl_multi_pw_aff *mpa)
 {
-	return isl_ast_build_from_multi_pw_aff(build, isl_ast_op_access, mpa);
+	return isl_ast_build_from_multi_pw_aff(build,
+						isl_ast_expr_op_access, mpa);
 }
 
 /* Construct an isl_ast_expr of type "type" that calls or accesses
@@ -2436,7 +2554,7 @@ __isl_give isl_ast_expr *isl_ast_build_access_from_multi_pw_aff(
  * The domain of "pma" is assumed to live in the external schedule domain.
  */
 static __isl_give isl_ast_expr *isl_ast_build_from_pw_multi_aff(
-	__isl_keep isl_ast_build *build, enum isl_ast_op_type type,
+	__isl_keep isl_ast_build *build, enum isl_ast_expr_op_type type,
 	__isl_take isl_pw_multi_aff *pma)
 {
 	isl_multi_pw_aff *mpa;
@@ -2454,7 +2572,8 @@ static __isl_give isl_ast_expr *isl_ast_build_from_pw_multi_aff(
 __isl_give isl_ast_expr *isl_ast_build_call_from_pw_multi_aff(
 	__isl_keep isl_ast_build *build, __isl_take isl_pw_multi_aff *pma)
 {
-	return isl_ast_build_from_pw_multi_aff(build, isl_ast_op_call, pma);
+	return isl_ast_build_from_pw_multi_aff(build,
+						isl_ast_expr_op_call, pma);
 }
 
 /* Construct an isl_ast_expr that accesses the array element specified by "pma".
@@ -2466,7 +2585,8 @@ __isl_give isl_ast_expr *isl_ast_build_call_from_pw_multi_aff(
 __isl_give isl_ast_expr *isl_ast_build_access_from_pw_multi_aff(
 	__isl_keep isl_ast_build *build, __isl_take isl_pw_multi_aff *pma)
 {
-	return isl_ast_build_from_pw_multi_aff(build, isl_ast_op_access, pma);
+	return isl_ast_build_from_pw_multi_aff(build,
+						isl_ast_expr_op_access, pma);
 }
 
 /* Construct an isl_ast_expr that calls the domain element
@@ -2485,7 +2605,7 @@ __isl_give isl_ast_node *isl_ast_build_call_from_executed(
 	iteration = isl_ast_build_compute_gist_pw_multi_aff(build, iteration);
 	iteration = isl_pw_multi_aff_intersect_domain(iteration,
 					isl_ast_build_get_domain(build));
-	expr = isl_ast_build_from_pw_multi_aff_internal(build, isl_ast_op_call,
-							iteration);
+	expr = isl_ast_build_from_pw_multi_aff_internal(build,
+					isl_ast_expr_op_call, iteration);
 	return isl_ast_node_alloc_user(expr);
 }
